from __future__ import annotations
import contextlib
from typing import Union, TYPE_CHECKING

from openqasm import ast

from oqpy.classical_types import (
    AstConvertible,
    BitVar,
    convert_range,
    map_to_ast,
    to_ast,
)

if TYPE_CHECKING:
    from oqpy.program import Program


__all__ = ["Qubit", "gate", "QubitArray", "defcal"]


class Qubit:
    def __init__(
        self,
        program: Program,
        identifier: Union[str, ast.Identifier, ast.IndexIdentifier],
        autodeclare=True,
    ):
        self.program = program
        if isinstance(identifier, str):
            self.ident = ast.Identifier(identifier)
            if autodeclare:
                program.add_statement(ast.QubitDeclaration(self.ident, None))
        else:
            assert not autodeclare
            self.ident = identifier

    def define_measure(self):
        pass

    def measure(self, output_location: BitVar):
        self.program.add_statement(
            ast.QuantumMeasurementAssignment(self.ident, output_location.to_ast())
        )


class _PhysicalQubits:
    def __init__(self, program: Program):
        self.program = program

    def __getitem__(self, item: int) -> Qubit:
        return Qubit(self.program, ast.Identifier(f"${item}"), autodeclare=False)


def gate(program: Program, qubits: Union[Qubit, list[Qubit]], name: str, *args):
    if isinstance(qubits, Qubit):
        qubits = [qubits]
    program.add_statement(
        ast.QuantumGate(
            [],
            ast.Identifier(name),
            map_to_ast(args),
            [q.ident for q in qubits],
        )
    )


class QubitArray:
    def __init__(
        self,
        program: Program,
        identifier: Union[str, ast.Identifier, ast.IndexIdentifier],
        size: int,
        autodeclare=True,
    ):
        self.program = program
        self.size = size
        if isinstance(identifier, str):
            self.ident = ast.Identifier(identifier)
            if autodeclare:
                program.add_statement(ast.QubitDeclaration(self.ident, ast.IntegerLiteral(size)))
        else:
            assert not autodeclare
            self.ident = identifier

    def measure(self, output_location: BitVar):
        self.program.add_statement(
            ast.QuantumMeasurementAssignment(
                ast.QuantumMeasurement(self.ident),
                output_location.to_ast(),
            )
        )

    def __getitem__(self, item):
        name = self.ident.name
        if isinstance(item, (list, tuple)):
            return QubitArray(ast.Selection(name, map_to_ast(item)), len(item), autodeclare=False)
        if isinstance(item, slice):
            return QubitArray(
                ast.Slice(name, convert_range(item)), item.stop - item.start, autodeclare=False
            )
        if isinstance(item, AstConvertible.__args__):
            return Qubit(ast.Subscript(name, to_ast(item)), autodeclare=False)
        raise TypeError


@contextlib.contextmanager
def defcal(program: Program, qubits: Union[Qubit, list[Qubit]], name: str):
    program.push()
    yield
    state = program.pop()

    if isinstance(qubits, Qubit):
        qubits = [qubits]

    program.add_statement(
        ast.CalibrationDefinition(
            ast.Identifier(name),
            [],  # todo: support arguments
            [q.ident for q in qubits],
            None,  # todo: support return type,
            state.body,
        )
    )



from __future__ import annotations
import functools
import inspect
from typing import Callable, get_type_hints, TYPE_CHECKING

from openqasm import ast

from oqpy.classical_types import (
    AstConvertible,
    OQFunctionCall,
    OQPyExpression,
    _Var,
    map_to_ast,
    to_ast,
)
from oqpy.quantum_types import Qubit, QubitArray

if TYPE_CHECKING:
    from oqpy.program import Program


__all__ = ["subroutine", "declare_extern"]


def subroutine(func):
    @functools.wraps(func)
    def wrapper(program, *args):
        name = func.__name__
        identifier = ast.Identifier(func.__name__)
        argnames = inspect.signature(func).parameters.keys()
        type_hints = get_type_hints(func)
        inputs = {}  # used as inputs when calling the actual python function
        arguments = []  # used in the ast definition of the subroutine
        for argname in argnames:
            if argname not in type_hints:
                raise ValueError(f"No type hint provided for {argname} on subroutine {name}")
            input_ = inputs[argname] = type_hints[argname](
                program, ident=argname, autodeclare=False
            )
            input_.set_captured()

            if isinstance(input_, _Var):
                arguments.append(ast.ClassicalArgument(input_.type, ast.Identifier(argname)))
            elif isinstance(input_, Qubit):
                arguments.append(ast.QuantumArgument(input_.ident, None))
            elif isinstance(input_, QubitArray):
                arguments.append(ast.QuantumArgument(input_.ident, ast.IntegerLiteral(input_.size)))
            else:
                raise ValueError(
                    f"Type hint for {argname} on subroutine {name} is not an oqpy variable type"
                )

        program.push()
        output = func(**inputs)
        state = program.pop()
        body = state.body
        if isinstance(output, OQPyExpression):
            return_type = output.type
            body.append(ast.ReturnStatement(to_ast(output)))
        elif output is None:
            return_type = None
        else:
            raise ValueError(
                "Output type of subroutine {name} was neither oqpy expression nor None"
            )
        stmt = ast.SubroutineDefinition(
            identifier,
            arguments=arguments,
            return_type=return_type,
            body=body,
        )

        program.add_subroutine(name, stmt)
        return OQFunctionCall(program, identifier, args, return_type)

    return wrapper


def declare_extern(program: Program, name: str, argtypes, return_type) -> Callable:
    extern_decl = ast.ExternDeclaration(
        ast.Identifier(name),
        argtypes,
        return_type,
    )

    def call_extern(*args: list[AstConvertible]):
        return OQFunctionCall(program, name, map_to_ast(args), return_type, extern_decl=extern_decl)

    return call_extern



from __future__ import annotations
import functools
import random
import string
from typing import Optional, Type, Union, final, TYPE_CHECKING

from openqasm import ast

if TYPE_CHECKING:
    from oqpy.program import Program


__all__ = [
    "AstConvertible",
    "IntVar",
    "FloatVar",
    "BitVar",
    "ComplexVar",
    "DurationVar",
    "Stretch",
]


class OQPyExpression:
    def __init__(self, program: Program):
        self.program = program
        self.captured = False

    def type(self):
        raise NotImplementedError

    def set_captured(self):
        self.captured = True

    @final
    def to_ast(self):
        self.set_captured()
        return self._to_ast()

    def _to_ast(self) -> ast.Expression:
        raise NotImplementedError

    def to_binary(self, op_name: str, first, second) -> OQPyBinaryExpression:
        return OQPyBinaryExpression(self.program, ast.BinaryOperator[op_name], first, second)

    def __add__(self, other):
        return self.to_binary("+", self, other)

    def __radd__(self, other):
        return self.to_binary("+", other, self)

    def __mod__(self, other):
        return self.to_binary("%", self, other)

    def __rmod__(self, other):
        return self.to_binary("%", other, self)

    def __mul__(self, other):
        return self.to_binary("*", self, other)

    def __rmul__(self, other):
        return self.to_binary("*", other, self)

    def __eq__(self, other):
        return self.to_binary("==", self, other)

    def __ne__(self, other):
        return self.to_binary("!=", self, other)

    def __gt__(self, other):
        return self.to_binary(">", self, other)

    def __lt__(self, other):
        return self.to_binary("<", self, other)

    def __ge__(self, other):
        return self.to_binary(">=", self, other)

    def __le__(self, other):
        return self.to_binary("<=", self, other)

    def __del__(self):
        if not self.captured:
            self.program.add_statement(ast.ExpressionStatement(self.to_ast()))


AstConvertible = Union[OQPyExpression, int, float, ast.Expression]


def convert_range(item: Union[slice, range]) -> ast.RangeDefinition:
    return ast.RangeDefinition(to_ast(item.start), to_ast(item.stop - 1), to_ast(item.step))


def to_ast(item: AstConvertible) -> ast.Expression:
    if isinstance(item, int):
        return ast.IntegerLiteral(item)
    if isinstance(item, float):
        return ast.RealLiteral(item)
    if isinstance(item, OQPyExpression):
        return item.to_ast()
    if isinstance(item, ast.Expression):
        return item
    raise TypeError(f"Cannot convert {item} of type {type(item)} to ast")


def optional_ast(item: Optional[AstConvertible]) -> Optional[ast.Expression]:
    if item is None:
        return None
    return to_ast(item)


def map_to_ast(items: list[AstConvertible]) -> list[ast.Expression]:
    return list(map(to_ast, items))


class OQPyBinaryExpression(OQPyExpression):
    def __init__(
        self, program: Program, op: ast.BinaryOperator, lhs: AstConvertible, rhs: AstConvertible
    ):
        super().__init__(program)
        self.op = op
        self.lhs = lhs
        self.rhs = rhs
        # TODO: Fix this!
        if isinstance(lhs, OQPyExpression):
            self.type = lhs.type
        elif isinstance(rhs, OQPyExpression):
            self.type = rhs.type
        else:
            raise TypeError("Neither lhs nor rhs is an expression?")

    def _to_ast(self) -> ast.Expression:
        return ast.BinaryExpression(self.op, to_ast(self.lhs), to_ast(self.rhs))


def make_identifier_name() -> str:
    return "".join([random.choice(string.ascii_letters) for _ in range(10)])


class _Var(OQPyExpression):
    type_cls: Type[ast.ClassicalType]

    def __init__(
        self,
        program: Program,
        init_expression=None,
        ident: Optional[str] = None,
        autodeclare=True,
        **type_kwargs,
    ):
        super().__init__(program)
        self.ident = ast.Identifier(ident or make_identifier_name())
        self.type = self.type_cls(**type_kwargs)
        init_expression = optional_ast(init_expression)
        if autodeclare:
            self.set_captured()
            self.program.add_statement(
                ast.ClassicalDeclaration(self.type, self.ident, init_expression)
            )

    def _to_ast(self) -> ast.Identifier:
        return self.ident

    def _do_assignment(self, other, op: ast.AssignmentOperator):
        self.program.add_statement(ast.ClassicalAssignment(self.to_ast(), op, to_ast(other)))

    def set(self, other):
        self._do_assignment(other, ast.AssignmentOperator["="])

    def __iadd__(self, other):
        self._do_assignment(other, ast.AssignmentOperator["+="])
        return self

    def __isub__(self, other):
        self._do_assignment(other, ast.AssignmentOperator["-="])
        return self

    # def __getitem__(self, item: int):
    #     return VarSlice(self, item)


class _SizedVar(_Var):
    default_size = None

    def __class_getitem__(cls, item: int):
        return functools.partial(cls, size=item)

    def __init__(self, *args, size=None, **kwargs):
        if size is None:
            size = self.default_size
        if size is not None:
            size = to_ast(size)
        super().__init__(*args, **kwargs, size=size)


class IntVar(_SizedVar):
    type_cls = ast.IntType
    default_size = 32

    def __imod__(self, other):
        self._do_assignment(other, ast.AssignmentOperator["%="])


class FloatVar(_SizedVar):
    type_cls = ast.FloatType
    default_size = 64


class BitVar(_SizedVar):
    type_cls = ast.BitType


class ComplexVar(_Var):
    type_cls = ast.ComplexType

    def __class_getitem__(cls, item):
        return functools.partial(cls, base_type=item)

    def __init__(self, program: Program, *args, base_type=FloatVar, **kwargs):
        base_type_instance = base_type(program, autodeclare=False)
        base_type_instance.set_captured()
        super().__init__(program, *args, **kwargs, base_type=base_type_instance.type)


def make_duration(time) -> ast.Expression:
    if isinstance(time, float):
        # Todo: make better units?
        return ast.DurationLiteral(1e9 * time, ast.TimeUnit.ns)
    elif isinstance(time, OQPyExpression):
        return to_ast(time)


class DurationVar(_Var):
    type_cls = ast.DurationType

    def __init__(
        self,
        program: Program,
        init_expression=None,
        identifier: str = None,
        autodeclare=True,
        **type_kwargs,
    ):
        if init_expression is not None:
            init_expression = make_duration(init_expression)
        super().__init__(program, init_expression, identifier, autodeclare, **type_kwargs)


class Stretch(_Var):
    type_cls = ast.StretchType


class OQFunctionCall(OQPyExpression):
    def __init__(
        self,
        program: Program,
        identifier: Union[str, ast.Identifier],
        args,
        return_type,
        extern_decl=None,
    ):
        super().__init__(program)
        if isinstance(identifier, str):
            identifier = ast.Identifier(identifier)
        self.identifier = identifier
        self.args = args
        self.type = return_type
        self.extern_decl = extern_decl

    def _to_ast(self):
        if self.extern_decl is not None:
            self.program.externs[self.identifier.name] = self.extern_decl
        return ast.FunctionCall(self.identifier, map_to_ast(self.args))



from __future__ import annotations
import contextlib
from typing import TYPE_CHECKING

from openqasm import ast

from oqpy.classical_types import make_duration

if TYPE_CHECKING:
    from oqpy.program import Program


__all__ = ["barrier", "delay", "Box"]


def barrier(program: Program, qubits=()):
    program.add_statement(ast.QuantumBarrier([q.ident for q in qubits]))


def delay(program: Program, time, qubits=()):
    program.add_statement(ast.DelayInstruction([], make_duration(time), [q.ident for q in qubits]))


@contextlib.contextmanager
def Box(program: Program, duration=None):
    if duration is not None:
        duration = make_duration(duration)
    program.push()
    yield
    state = program.pop()
    program.add_statement(ast.Box(duration, state.body))



from __future__ import annotations
import contextlib
from typing import Optional, TYPE_CHECKING

from openqasm import ast

from oqpy.classical_types import IntVar, OQPyExpression, convert_range, map_to_ast, to_ast

if TYPE_CHECKING:
    from oqpy.program import Program


__all__ = ["If", "Else", "ForIn", "While"]


@contextlib.contextmanager
def If(program: Program, condition: OQPyExpression):
    program.push()
    yield
    state = program.pop()
    program.state.add_if_clause(to_ast(condition), state.body)


@contextlib.contextmanager
def Else(program: Program):
    program.push()
    yield
    state = program.pop()
    program.state.add_else_clause(state.body)


@contextlib.contextmanager
def ForIn(program: Program, iterator, identifier: Optional[str] = None):
    program.push()
    var = IntVar(program, ident=identifier, autodeclare=False)
    var.set_captured()
    yield var
    state = program.pop()

    if isinstance(iterator, list):
        iterator = map_to_ast(iterator)
    elif isinstance(iterator, range):
        iterator = convert_range(iterator)
    else:
        iterator = to_ast(iterator)

    stmt = ast.ForInLoop(var.ident, iterator, state.body)
    program.add_statement(stmt)


@contextlib.contextmanager
def While(program: Program, condition: OQPyExpression):
    program.push()
    yield
    state = program.pop()
    program.add_statement(ast.WhileLoop(to_ast(condition), state.body))



from __future__ import annotations
import contextlib
from typing import Optional, TYPE_CHECKING

from openqasm import ast

from oqpy.classical_types import AstConvertible, OQFunctionCall, _Var
from oqpy.subroutines import declare_extern

if TYPE_CHECKING:
    from oqpy.program import Program


__all__ = ["PortVar", "WaveformVar", "Frame", "constant_waveform", "gaussian_waveform", "Cal"]


class PortVar(_Var):
    type_cls = ast.PulseType

    def __init__(self, program: Program, ident=None, **kwargs):
        super().__init__(program, ident=ident, **kwargs, type=ast.PulseTypeName.port)


class WaveformVar(_Var):
    type_cls = ast.PulseType

    def __init__(self, program: Program, init_expression=None, ident=None, **kwargs):
        super().__init__(
            program,
            init_expression=init_expression,
            ident=ident,
            **kwargs,
            type=ast.PulseTypeName.waveform,
        )


class Frame(_Var):
    type_cls = ast.PulseType

    def __init__(
        self,
        program: Program = None,
        port: Optional[PortVar] = None,
        frequency: AstConvertible = None,
        phase: AstConvertible = 0,
        ident=None,
        autodeclare=True,
    ):
        if port is None != frequency is None:
            raise ValueError("Must declare both port and frequency or neither")
        init_expression = port and OQFunctionCall(
            program, "newframe", [port, frequency, phase], ast.PulseType(ast.PulseTypeName.frame)
        )
        super().__init__(
            program, init_expression, ident, autodeclare=autodeclare, type=ast.PulseTypeName.frame
        )

    def _resolve_program(self, program: Program | None) -> Program:
        if program is not None:
            return program
        if self.program is None

    def play(self, waveform: AstConvertible, program):
        self.set_captured()
        program = self._resolve_program
        return OQFunctionCall(self.program, "play", [waveform, self.ident], None)

    def capture(self, kernel: AstConvertible):
        self.set_captured()
        return OQFunctionCall(self.program, "capture", [kernel, self.ident], None)

    def set_phase(self, phase: AstConvertible):
        self.set_captured()
        return OQFunctionCall(self.program, "set_phase", [phase, self.ident], None)

    def shift_phase(self, phase: AstConvertible):
        self.set_captured()
        return OQFunctionCall(self.program, "shift_phase", [phase, self.ident], None)

    def set_frequency(self, freq: AstConvertible):
        self.set_captured()
        return OQFunctionCall(self.program, "set_frequency", [freq, self.ident], None)

    def shift_frequency(self, freq: AstConvertible):
        self.set_captured()
        return OQFunctionCall(self.program, "shift_frequency", [freq, self.ident], None)

    def set_scale(self, scale: AstConvertible):
        self.set_captured()
        return OQFunctionCall(self.program, "set_scale", [scale, self.ident], None)

    def shift_scale(self, scale: AstConvertible):
        self.set_captured()
        return OQFunctionCall(self.program, "shift_scale", [scale, self.ident], None)


def _make_waveform(program: Program, name: str, argtypes, args):
    func = declare_extern(program, name, argtypes, ast.PulseType(ast.PulseTypeName.waveform))
    return func(*args)


def constant_waveform(program: Program, length: float, amp: float):
    return _make_waveform(
        program,
        "constant",
        [ast.DurationType(), ast.FloatType(ast.IntegerLiteral(64))],
        [length, amp],
    )


def gaussian_waveform(
    program: Program, length: float, sigma: float, amp: float, phase: float = 0.0
):
    return _make_waveform(
        program,
        "gaussian",
        [
            ast.DurationType(),
            ast.DurationType(),
            ast.FloatType(ast.IntegerLiteral(64)),
            ast.FloatType(ast.IntegerLiteral(64)),
        ],
        [length, sigma, amp, phase],
    )


@contextlib.contextmanager
def Cal(program):
    """Context manager that begins a cal block."""
    program.push()
    yield
    state = program.pop()
    program.add_statement(ast.CalibrationBlock(state.body))



__version__ = "0.2.0"
from oqpy.classical_types import *
from oqpy.control_flow import *
from oqpy.program import *
from oqpy.pulse import *
from oqpy.timing import *
from oqpy.quantum_types import *
from oqpy.subroutines import *



from __future__ import annotations

from copy import deepcopy
from typing import Optional, TYPE_CHECKING

from openqasm import ast
from openqasm.printer import print_qasm
from oqpy import pulse
from oqpy import timing
from oqpy import quantum_types

if TYPE_CHECKING:
    from oqpy import AstConvertible


__all__ = ["Program"]


class ProgramState:
    def __init__(self):
        self.body: list[ast.Statement] = []
        self.if_clause: Optional[ast.BranchingStatement] = None

    def add_if_clause(self, condition: ast.Expression, if_clause: list[ast.Statement]):
        self.finalize_if_clause()
        self.if_clause = ast.BranchingStatement(condition, if_clause, [])

    def add_else_clause(self, else_clause: list[ast.Statement]):
        if self.if_clause is None:
            raise RuntimeError("Else without If.")
        self.if_clause.else_block = else_clause
        self.finalize_if_clause()

    def finalize_if_clause(self):
        if self.if_clause is not None:
            if_clause, self.if_clause = self.if_clause, None
            self.add_statement(if_clause)

    def add_statement(self, stmt: ast.Statement):
        self.finalize_if_clause()
        self.body.append(stmt)


class Program:
    def __init__(self):
        self.stack: list[ProgramState] = [ProgramState()]
        self.defcals: dict[tuple[str, str], ast.CalibrationDefinition] = {}
        self.subroutines: dict[str, ast.SubroutineDefinition] = {}
        self.externs: dict[str, ast.ExternDeclaration] = {}
        self.physical_qubits = quantum_types._PhysicalQubits(self)

    def __iadd__(self, other):
        if len(other.stack) > 1:
            raise RuntimeError("Cannot add subprogram with unclosed contextmanagers")
        self.state.finalize_if_clause()
        self.state.body.extend(other.state.body)
        self.state.if_clause = other.state.if_clause
        self.state.finalize_if_clause()
        self.defcals.update(other.defcals)
        self.subroutines.update(other.subroutines)
        self.externs.update(other.externs)
        return self

    def __add__(self, other):
        self_copy = deepcopy(self)
        self_copy += other
        return self_copy

    @property
    def state(self) -> ProgramState:
        return self.stack[-1]

    def push(self):
        self.stack.append(ProgramState())

    def pop(self) -> ProgramState:
        state = self.stack.pop()
        state.finalize_if_clause()
        return state

    def add_statement(self, stmt):
        self.state.add_statement(stmt)

    def add_subroutine(self, name: str, stmt: ast.SubroutineDefinition):
        self.subroutines[name] = stmt

    def add_defcal(self, qubit_name: str, name: str, stmt: ast.CalibrationDefinition):
        self.defcals[(qubit_name, name)] = stmt

    def to_ast(self, encal=False, include_externs=True) -> ast.Program:
        assert len(self.stack) == 1
        self.state.finalize_if_clause()
        statements = []
        if include_externs:
            statements += list(self.externs.values())
        statements += list(self.subroutines.values()) + self.state.body
        if encal:
            statements = [ast.CalibrationBlock(statements)]
        return ast.Program(statements=statements)

    def to_qasm(self, encal=False, include_externs=True) -> str:
        return print_qasm(
            self.to_ast(encal=encal, include_externs=include_externs), pretty_print=True
        )

    def delay(self, time, qubits=()):
        timing.delay(self, time, qubits)
        return self

    def barrier(self, qubits):
        timing.barrier(self, qubits)
        return self

    def play(self, frame: pulse.Frame, waveform: AstConvertible):
        frame.play(waveform)
        return self

    def capture(self, frame: pulse.Frame, kernel: AstConvertible):
        frame.capture(kernel)
        return self

    def set_phase(self, frame: pulse.Frame, phase: AstConvertible):
        frame.set_phase(phase)
        return self

    def shift_phase(self, frame: pulse.Frame, phase: AstConvertible):
        frame.shift_phase(phase)
        return self

    def set_frequency(self, frame: pulse.Frame, freq: AstConvertible):
        frame.set_frequency(freq)
        return self

    def shift_frequency(self, frame: pulse.Frame, freq: AstConvertible):
        frame.shift_frequency(freq)
        return self

    def set_scale(self, frame: pulse.Frame, scale: AstConvertible):
        frame.set_scale(scale)
        return self

    def shift_scale(self, frame: pulse.Frame, scale: AstConvertible):
        frame.shift_scale(scale)
        return self

    def gate(self, qubits, name):
        quantum_types.gate(self, qubits, name)
        return self



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['oqpy']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.20.2,<2.0.0',
 'openpulse-poc @ vendor/openpulse_poc-0.0.9-py3-none-any.whl',
 'pydantic>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'oqpy',
    'version': '0.2.0',
    'description': '',
    'long_description': None,
    'author': 'Philip Reinhold',
    'author_email': 'pcrein@amazon.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)
